import os
import random
import time

# Set Qt API to PySide6 before importing matplotlib
os.environ.setdefault("QT_API", "PySide6")

import matplotlib

# Set matplotlib backend to PySide6-compatible qtagg
matplotlib.use("qtagg")
import matplotlib.pyplot as plt
import numpy as np

# Import from the PySide6-compatible backend
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib.figure import Figure
from PySide6 import QtCore, QtWidgets
from PySide6.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget

from xpcsviewer.utils.logging_config import get_logger

from .plot_constants import MATPLOTLIB_COLORS_HEX as colors
from .plot_constants import get_color_marker

logger = get_logger(__name__)

# hide the lines in legend
# https://stackoverflow.com/questions/21285885
# matplotlib.rcParams['legend.handlelength'] = 2
# matplotlib.rcParams['legend.numpoints'] = 1

# matplotlib.pyplot.style.use(['science', 'no-latex'])


class NavigationToolbarSimple(NavigationToolbar2QT):
    def __init__(self, *kw, **kwargs):
        super().__init__(*kw, **kwargs)

    def mouse_move(self, event):
        # just disable the mose_move event
        pass


class MplCanvasBarH(QtWidgets.QWidget):
    """
    MplWidget combines a MplCanvas with a vertical toolbar
    """

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.hdl = MplCanvas()
        self.navi_toolbar = NavigationToolbarSimple(self.hdl, self)
        self.navi_toolbar.setOrientation(QtCore.Qt.Vertical)
        self.hbl = QHBoxLayout()
        self.hbl.addWidget(self.hdl)
        self.hbl.addWidget(self.navi_toolbar)
        # self.navi_toolbar.setOrientation(QtCore.Qt.Vertical)
        self.setLayout(self.hbl)

    def clear(self):
        self.hdl.clear()
        self.hdl.draw()

    def apply_theme(self, theme: str) -> None:
        """Apply theme colors to the embedded canvas."""
        self.hdl.apply_theme(theme)


class MplCanvasBarV(QWidget):
    """
    MplWidget combines a MplCanvas with a horizontal toolbar
    """

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.hdl = MplCanvas()
        self.navi_toolbar = NavigationToolbar2QT(self.hdl, self)
        self.vbl = QVBoxLayout()
        self.vbl.addWidget(self.navi_toolbar)
        self.vbl.addWidget(self.hdl)
        self.setLayout(self.vbl)

    def clear(self):
        self.hdl.clear()
        self.hdl.draw()

    def apply_theme(self, theme: str) -> None:
        """Apply theme colors to the embedded canvas."""
        self.hdl.apply_theme(theme)


class MplCanvasBar(QWidget):
    """
    MplWidget combines a MplCanvas with a Toolbar
    """

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.hdl = MplCanvas()
        self.navi_toolbar = NavigationToolbarSimple(self.hdl, self)
        self.vbl = QVBoxLayout()
        self.vbl.addWidget(self.hdl)
        self.vbl.addWidget(self.navi_toolbar)
        # self.navi_toolbar.setOrientation(QtCore.Qt.Vertical)
        self.setLayout(self.vbl)


class MplCanvas(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=15, height=12, dpi=100):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        # self.axes = fig.add_subplot(111)
        super().__init__(self.fig)
        self.shape = None
        self.axes = None
        self.obj = None
        self.line_builder = None
        self.cids = []

    def apply_theme(self, theme: str) -> None:
        """
        Apply theme colors to this matplotlib canvas.

        Parameters
        ----------
        theme : str
            Either "light" or "dark"
        """
        from .plot_constants import get_theme_colors

        colors = get_theme_colors(theme)

        # Update figure background
        self.fig.set_facecolor(colors["background"])

        # Update axes if they exist
        if self.axes is not None:
            axes_list = (
                np.array(self.axes).ravel() if self.shape != (1, 1) else [self.axes]
            )
            for ax in axes_list:
                ax.set_facecolor(colors["background"])
                ax.tick_params(colors=colors["axis"])
                ax.xaxis.label.set_color(colors["text"])
                ax.yaxis.label.set_color(colors["text"])
                ax.title.set_color(colors["text"])
                for spine in ax.spines.values():
                    spine.set_edgecolor(colors["axis"])

        self.draw()

    def link_line_builder(self, lb_type=None):
        if lb_type is None:
            self.unlink_line_builder()

        if self.line_builder is not None and self.line_builder.lb_type != lb_type:
            self.unlink_line_builder()

        if self.line_builder is None and self.shape == (1, 1):
            self.line_builder = LineBuilder(self.fig, self.axes, lb_type)
            cid1 = self.fig.canvas.mpl_connect(
                "button_press_event", self.line_builder.mouse_click
            )
            cid2 = self.fig.canvas.mpl_connect(
                "motion_notify_event", self.line_builder.mouse_move
            )
            self.cids = [cid1, cid2]

    def unlink_line_builder(self):
        if self.line_builder is not None:
            for cid in self.cids:
                self.fig.canvas.mpl_disconnect(cid)
            self.line_builder.clear()
            self.line_builder = None

    def subplots(self, n, m, **kwargs):
        self.axes = self.fig.subplots(n, m, **kwargs)
        self.shape = (n, m)
        return self.axes

    def clear(self):
        self.unlink_line_builder()
        self.clear_axes()
        self.fig.clear()
        self.axes = None
        self.obj = None

    def adjust_canvas_size(self, num_col, num_row):
        t = self.parent().parent().parent()
        aspect = 1 / 1.618 if t is None else t.height() / self.width()

        min_size = t.height() - 20
        width = self.width()
        canvas_size = max(min_size, int(width / num_col * aspect * num_row))
        self.setMinimumSize(QtCore.QSize(0, canvas_size))

    def clear_axes(self):
        if self.axes is None:
            return
        if self.shape == (1, 1):
            self.axes.clear()
        else:
            for ax in self.axes.ravel():
                ax.clear()

    def auto_scale(self, ylim=None, xlim=None, xscale=None, yscale=None):
        if self.axes is None:
            return
        for ax in np.array(self.axes).ravel():
            if xscale is not None:
                ax.set_xscale(xscale)
            if yscale is not None:
                ax.set_yscale(yscale)
            ax.relim()
            ax.autoscale_view(True, True, True)
            if ylim is not None:
                ax.set_ylim(ylim)
            if xlim is not None:
                ax.set_xlim(xlim)

    def update_lin(self, loc, x, y, visible=True):
        if self.obj is None:
            return
        (lin_obj,) = self.obj["lin"][loc]
        lin_obj.set_data(x, y)
        lin_obj.set_visible(visible)
        return

    def update_err(self, loc, x, y, y_error):
        if self.obj is None:
            return
        err_obj = self.obj["err"][loc]
        adjust_yerr(err_obj, x, y, y_error)
        return

    def show_image(
        self,
        data,
        vmin=None,
        vmax=None,
        extent=None,
        cmap="seismic",
        xlabel=None,
        ylabel=None,
        title=None,
        id_list=None,
        vline_freq=-1,
    ):
        def add_vline(ax, stop, vline_freq):
            if vline_freq < 0:
                return
            for x in np.arange(vline_freq, stop - 1, vline_freq):
                # for x in np.arange(1, stop // vline_freq - 1):
                ax.axvline(x - 0.5, ls="--", lw=0.5, color="black", alpha=0.5)

        if self.axes is None:
            ax = self.subplots(1, 1)
            add_vline(ax, data.shape[1], vline_freq)
            im0 = ax.imshow(
                data,
                aspect="auto",
                cmap=plt.get_cmap(cmap),
                vmin=vmin,
                vmax=vmax,
                extent=extent,
                interpolation=None,
            )
            self.fig.colorbar(im0, ax=ax)

            ax.set_title(title)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

            # when there are too many points, avoid labeling.
            # if data.shape < 20:
            #     ax.set_xticks(np.arange(data.shape[1]))
            #     ax.set_xticklabels(id_list[0: data.shape[1]])

            self.obj = [im0]
            self.fig.tight_layout()
        else:
            self.obj[0].set_data(data)
            self.obj[0].set_clim(vmin, vmax)
            self.axes.set_title(title)
            self.axes.set_xlabel(xlabel)
            self.axes.set_ylabel(ylabel)

        self.draw()

    def show_lines(
        self,
        data,
        xlabel=None,
        ylabel=None,
        title=None,
        legend=None,
        loc="best",
        rows=None,
        marker_size=3,
    ):
        if legend in [None, False]:
            legend = np.arange(len(data))

        if rows in [None, []]:
            alpha = np.ones(len(data)) * 0.75
        else:
            alpha = np.ones(len(data)) * 0.15
            for t in rows:
                if t < len(data):
                    alpha[t] = 1.0

        if isinstance(data, np.ndarray):
            x = np.arange(data.shape[1])
            data2 = []
            for n in range(data.shape[0]):
                data2.append([x, data[n]])
            data = data2

        if self.axes is None or len(data) != len(self.obj):
            ax = self.subplots(1, 1)
            line_obj = []
            for n in range(len(data)):
                cl, mk = get_color_marker(n)
                line = ax.plot(
                    data[n][0],
                    data[n][1],
                    mk + "-",
                    ms=marker_size,
                    alpha=alpha[n],
                    label=legend[n],
                    color=cl,
                    mfc="none",
                )
                line_obj.append(line)
            self.obj = line_obj

            if legend is not None and loc != "outside":
                ax.legend(loc=loc)
            elif loc == "outside":
                ax.legend(bbox_to_anchor=(1.03, 1.0), loc="upper left")

        else:
            for n in range(len(data)):
                (line,) = self.obj[n]
                line.set_data(data[n][0], data[n][1])
                if legend is not None:
                    line.set_label(legend[n])
            self.auto_scale()

        self.axes.set_title(title)
        self.axes.set_xlabel(xlabel)
        self.axes.set_ylabel(ylabel)
        self.fig.tight_layout(rect=(0.07, 0.07, 0.93, 0.93))
        self.draw()

    def show_scatter(
        self,
        data,
        color=None,
        xlabel=None,
        ylabel=None,
        title=None,
        legend=None,
        loc="best",
        alpha=0.85,
    ):
        if data.ndim != 2 or data.shape[0] != 2:
            raise ValueError("input data shape not supported")
        x, y = data[0], data[1]
        if color is None:
            color = np.arange(x.size)
        # if legend in [None, False]:
        #     legend = np.arange(len(x))

        if self.axes is not None:
            self.clear()
        ax = self.subplots(1, 1)
        line = ax.scatter(x, y, c=color)
        self.fig.colorbar(line, ax=ax)
        if legend is not None:
            ax.legend(loc=loc)

        self.axes.set_title(title)
        self.axes.set_xlabel(xlabel)
        self.axes.set_ylabel(ylabel)
        self.draw()


# https://github.com/matplotlib/matplotlib/issues/4556
def adjust_yerr(err_obj, x, y, y_error):
    # not using error top / bot bar;
    # ln, (err_top, err_bot), (bars, ) = err_obj
    ln, _, (bars,) = err_obj
    ln.set_data(x, y)

    yerr_top = y + y_error
    yerr_bot = y - y_error

    # err_top.set_ydata(yerr_top)
    # err_bot.set_ydata(yerr_bot)

    new_segments = [
        np.array([[x, yt], [x, yb]])
        for x, yt, yb in zip(x, yerr_top, yerr_bot, strict=False)
    ]

    bars.set_segments(new_segments)


MplToolbar = NavigationToolbar2QT
# class MplToolbar(MplCanvas):
#     def __init__(self, **kwargs):
#         super(MplCanvas, self).__init__(**kwargs)
#         toolbar = NavigationToolbar2QT(self.fig, self)
#         layout = QtWidgets.QVBoxLayout()
#         layout.addWidget(toolbar)
#         layout.addWidget(self)
#
#         widget = QtWidgets.QWidget()
#         widget.setLayout(layout)
#         self.setCentralWidget(widget)


class LineBuilder:
    """
        code copied from
        http://chuanshuoge2.blogspot.com/2019/12/matplotlib-mouse-click-event\
            -draw-line.html

    """

    def __init__(self, fig, ax, lb_type="hline"):
        self.xs = []
        self.ys = []
        self.ax = ax
        self.fig = fig
        self.color = random.choice(colors)  # noqa: S311 - non-cryptographic UI randomness
        self.color_hist = []
        self.num_lines = 0
        self.labels = []
        self.curr_time = -1
        self.lb_type = lb_type
        self.reserve_lines = len(self.ax.lines)

    def clear(self):
        self.xs = []
        self.ys = []
        self.color_hist = []
        for _n in range(len(self.ax.lines) - self.reserve_lines):
            self.ax.lines.pop()
        for _n in range(len(self.labels)):
            label = self.labels.pop()
            label.remove()
        self.fig.canvas.draw()

    def plot_line(self):
        if self.lb_type == "slope":
            xa, xb = self.xs[-2], self.xs[-1]
            ya, yb = self.ys[-2], self.ys[-1]
            dn_term = np.log(xa / xb)
            if dn_term == 0:
                dn_term = 1e-8
            slope = np.log(ya / yb) / dn_term
            txt = f"$q^{{{slope:.2f}}}$"

            # compute position to add label, notice the plot should be
            # logx-logy; slightly offset cen_x to make the label more clear
            cen_x = np.sqrt(xa * xb * 1.2)
            cen_y = np.sqrt(ya * yb * 1.2)

        elif self.lb_type == "hline":
            xa, xb = self.xs[-2], self.xs[-1]
            ya, yb = self.ys[-1], self.ys[-1]
            delta_x = 2 * np.pi / abs(xa - xb)
            txt = f"$\\Delta_x={{{delta_x:.1f}}}\\AA$"

            cen_x = np.sqrt(xa * xa)
            cen_y = np.sqrt(ya * yb * 0.3)
        else:
            return

        (line,) = self.ax.plot([xa, xb], [ya, yb], self.color)
        label = self.ax.annotate(txt, (cen_x, cen_y), color=self.color)
        self.labels.append(label)

        line.figure.canvas.draw()
        self.num_lines += 1
        self.color_hist.append(self.color)
        self.color = random.choice(colors)  # noqa: S311 - non-cryptographic UI randomness

    def mouse_click(self, event):
        if not event.inaxes:
            return

        # left click
        if event.button == 1:
            self.xs.append(event.xdata)
            self.ys.append(event.ydata)

            if self.lb_type == "hline":
                line = self.ax.axvline(event.xdata, color=self.color, ls=":")
                line.figure.canvas.draw()

            # add a line to plot if it has 2 points
            if len(self.xs) % 2 == 0 and len(self.xs) > 0:
                self.plot_line()

        # right click
        if event.button == 3:
            if len(self.xs) == 0:
                return
            self.xs.pop()
            self.ys.pop()
            if self.lb_type == "hline":
                self.ax.lines.pop()
            # delete last line drawn if the line is missing a point,
            # never delete the original stock plot
            if len(self.xs) % 2 == 1 and len(self.ax.lines) > self.reserve_lines:
                self.ax.lines.pop()
                self.num_lines -= 1
                self.color = self.color_hist.pop()
                label = self.labels.pop()
                label.remove()
            # refresh plot
            self.fig.canvas.draw()

    def mouse_move(self, event):
        if self.lb_type is None or not event.inaxes:
            return
        # draw temporary line from a single point to the mouse position
        # delete the temporary line when mouse move to another position
        if time.perf_counter() - self.curr_time < 0.1:
            return

        self.curr_time = time.perf_counter()
        if len(self.xs) % 2 == 1:
            if self.lb_type == "slope":
                (line,) = self.ax.plot(
                    [self.xs[-1], event.xdata], [self.ys[-1], event.ydata], self.color
                )
            else:
                (line,) = self.ax.plot(
                    [self.xs[-1], event.xdata], [event.ydata, event.ydata], self.color
                )
            line.figure.canvas.draw()
            self.ax.lines.pop()



"""Unit tests for theme design tokens."""

import pytest

from xpcsviewer.gui.theme.tokens import (
    DARK_COLORS,
    DARK_TOKENS,
    LIGHT_COLORS,
    LIGHT_TOKENS,
    SPACING_TOKENS,
    TYPOGRAPHY_TOKENS,
    ColorTokens,
    SpacingTokens,
    ThemeDefinition,
    TypographyTokens,
)


class TestColorTokens:
    """Tests for ColorTokens dataclass."""

    def test_light_colors_are_valid_hex(self):
        """All light theme colors should be valid hex strings."""
        for field_name in ColorTokens.__dataclass_fields__:
            value = getattr(LIGHT_COLORS, field_name)
            assert isinstance(value, str)
            assert value.startswith("#"), f"{field_name} should start with #"
            assert len(value) == 7, f"{field_name} should be 7 chars (#RRGGBB)"

    def test_dark_colors_are_valid_hex(self):
        """All dark theme colors should be valid hex strings."""
        for field_name in ColorTokens.__dataclass_fields__:
            value = getattr(DARK_COLORS, field_name)
            assert isinstance(value, str)
            assert value.startswith("#"), f"{field_name} should start with #"
            assert len(value) == 7, f"{field_name} should be 7 chars (#RRGGBB)"

    def test_color_tokens_are_immutable(self):
        """ColorTokens should be immutable (frozen dataclass)."""
        with pytest.raises(AttributeError):
            LIGHT_COLORS.background_primary = "#000000"

    def test_light_and_dark_are_different(self):
        """Light and dark themes should have different color values."""
        assert LIGHT_COLORS.background_primary != DARK_COLORS.background_primary
        assert LIGHT_COLORS.text_primary != DARK_COLORS.text_primary

    def test_wcag_contrast_light_text(self):
        """Light theme text on background should have sufficient contrast."""
        # Basic check: primary text should be dark on light background
        assert (
            LIGHT_COLORS.text_primary.lower() < LIGHT_COLORS.background_primary.lower()
        )

    def test_wcag_contrast_dark_text(self):
        """Dark theme text on background should have sufficient contrast."""
        # Basic check: primary text should be light on dark background
        assert DARK_COLORS.text_primary.lower() > DARK_COLORS.background_primary.lower()


class TestSpacingTokens:
    """Tests for SpacingTokens dataclass."""

    def test_spacing_follows_8px_grid(self):
        """Spacing values should follow 8px base grid."""
        assert SPACING_TOKENS.xs == 4  # 0.5x
        assert SPACING_TOKENS.sm == 8  # 1x
        assert SPACING_TOKENS.md == 16  # 2x
        assert SPACING_TOKENS.lg == 24  # 3x
        assert SPACING_TOKENS.xl == 32  # 4x
        assert SPACING_TOKENS.xxl == 48  # 6x

    def test_spacing_tokens_are_immutable(self):
        """SpacingTokens should be immutable."""
        with pytest.raises(AttributeError):
            SPACING_TOKENS.sm = 10

    def test_spacing_values_are_integers(self):
        """All spacing values should be integers."""
        for field_name in SpacingTokens.__dataclass_fields__:
            value = getattr(SPACING_TOKENS, field_name)
            assert isinstance(value, int), f"{field_name} should be int"

    def test_spacing_values_are_positive(self):
        """All spacing values should be positive."""
        for field_name in SpacingTokens.__dataclass_fields__:
            value = getattr(SPACING_TOKENS, field_name)
            assert value > 0, f"{field_name} should be positive"


class TestTypographyTokens:
    """Tests for TypographyTokens dataclass."""

    def test_default_font_family(self):
        """Default font family should use system fonts."""
        # Font stack should include system fonts for cross-platform compatibility
        font_family = TYPOGRAPHY_TOKENS.font_family
        # Check for at least one of the common system font names
        has_system_font = any(
            f in font_family
            for f in ["-apple-system", "BlinkMacSystemFont", "Segoe UI", "system-ui"]
        )
        assert has_system_font, f"Expected system font in: {font_family}"
        assert "sans-serif" in font_family

    def test_font_sizes_are_reasonable(self):
        """Font sizes should be in reasonable range (8-36pt)."""
        assert 8 <= TYPOGRAPHY_TOKENS.size_xs <= 12
        assert 10 <= TYPOGRAPHY_TOKENS.size_sm <= 14
        assert 12 <= TYPOGRAPHY_TOKENS.size_base <= 16
        assert 14 <= TYPOGRAPHY_TOKENS.size_lg <= 18
        assert 16 <= TYPOGRAPHY_TOKENS.size_xl <= 24
        assert 20 <= TYPOGRAPHY_TOKENS.size_xxl <= 36

    def test_font_sizes_are_ascending(self):
        """Font sizes should increase from xs to xxl."""
        sizes = [
            TYPOGRAPHY_TOKENS.size_xs,
            TYPOGRAPHY_TOKENS.size_sm,
            TYPOGRAPHY_TOKENS.size_base,
            TYPOGRAPHY_TOKENS.size_lg,
            TYPOGRAPHY_TOKENS.size_xl,
            TYPOGRAPHY_TOKENS.size_xxl,
        ]
        assert sizes == sorted(sizes), "Font sizes should be in ascending order"

    def test_font_weights_are_valid(self):
        """Font weights should be valid CSS font-weight values."""
        assert 100 <= TYPOGRAPHY_TOKENS.weight_normal <= 400
        assert 400 <= TYPOGRAPHY_TOKENS.weight_medium <= 600
        assert 600 <= TYPOGRAPHY_TOKENS.weight_bold <= 900


class TestThemeDefinition:
    """Tests for ThemeDefinition dataclass."""

    def test_light_theme_name(self):
        """Light theme should have correct name."""
        assert LIGHT_TOKENS.name == "light"
        assert LIGHT_TOKENS.display_name == "Light Mode"

    def test_dark_theme_name(self):
        """Dark theme should have correct name."""
        assert DARK_TOKENS.name == "dark"
        assert DARK_TOKENS.display_name == "Dark Mode"

    def test_themes_share_spacing(self):
        """Light and dark themes should share spacing tokens."""
        assert LIGHT_TOKENS.spacing == DARK_TOKENS.spacing

    def test_themes_share_typography(self):
        """Light and dark themes should share typography tokens."""
        assert LIGHT_TOKENS.typography == DARK_TOKENS.typography

    def test_theme_has_all_components(self):
        """Theme definition should have all required components."""
        for theme in [LIGHT_TOKENS, DARK_TOKENS]:
            assert isinstance(theme.colors, ColorTokens)
            assert isinstance(theme.spacing, SpacingTokens)
            assert isinstance(theme.typography, TypographyTokens)

    def test_theme_definitions_are_immutable(self):
        """ThemeDefinition should be immutable."""
        with pytest.raises(AttributeError):
            LIGHT_TOKENS.name = "modified"



"""Unit tests for ViewerKernel class.

This module provides comprehensive unit tests for the ViewerKernel class,
covering backend coordination, analysis orchestration, and file management.
"""

import weakref
from unittest.mock import Mock, patch

import pytest

from xpcsviewer.file_locator import FileLocator
from xpcsviewer.helper.listmodel import TableDataModel
from xpcsviewer.module.average_toolbox import AverageToolbox
from xpcsviewer.viewer_kernel import ViewerKernel


class TestViewerKernelInit:
    """Test suite for ViewerKernel initialization."""

    def test_init_basic(self, temp_dir):
        """Test basic ViewerKernel initialization."""
        kernel = ViewerKernel(temp_dir)

        # Check inheritance from FileLocator
        assert isinstance(kernel, FileLocator)

        # Check initialization attributes
        assert kernel.path == temp_dir
        assert kernel.statusbar is None
        assert kernel.meta is not None
        assert isinstance(kernel.avg_tb, AverageToolbox)
        assert isinstance(kernel.avg_worker, TableDataModel)
        assert kernel.avg_jid == 0
        assert isinstance(kernel.avg_worker_active, dict)

        # Check memory management attributes
        assert hasattr(kernel, "_current_dset_cache")
        assert isinstance(kernel._current_dset_cache, weakref.WeakValueDictionary)
        assert isinstance(kernel._plot_kwargs_record, dict)
        assert kernel._memory_cleanup_threshold == 0.8

    def test_init_with_statusbar(self, temp_dir):
        """Test ViewerKernel initialization with statusbar."""
        mock_statusbar = Mock()
        kernel = ViewerKernel(temp_dir, statusbar=mock_statusbar)

        assert kernel.statusbar is mock_statusbar
        assert kernel.path == temp_dir

    def test_init_calls_reset_meta(self, temp_dir):
        """Test that initialization calls reset_meta."""
        with patch.object(ViewerKernel, "reset_meta") as mock_reset:
            ViewerKernel(temp_dir)
            mock_reset.assert_called_once()


class TestViewerKernelMetaManagement:
    """Test suite for ViewerKernel metadata management."""

    def test_reset_meta_structure(self, temp_dir):
        """Test reset_meta creates correct metadata structure."""
        kernel = ViewerKernel(temp_dir)
        kernel.reset_meta()

        # Check essential meta keys exist
        expected_keys = [
            "saxs1d_bkg_fname",
            "saxs1d_bkg_xf",
        ]

        for key in expected_keys:
            assert key in kernel.meta

        # Check initial values
        assert kernel.meta["saxs1d_bkg_fname"] is None
        assert kernel.meta["saxs1d_bkg_xf"] is None

    def test_reset_meta_preserves_type(self, temp_dir):
        """Test that reset_meta maintains dictionary type."""
        kernel = ViewerKernel(temp_dir)

        # Modify meta
        kernel.meta["custom_key"] = "custom_value"

        # Reset should create new dict
        kernel.reset_meta()

        assert isinstance(kernel.meta, dict)
        assert "custom_key" not in kernel.meta
        assert kernel.meta["saxs1d_bkg_fname"] is None


class TestViewerKernelMemoryManagement:
    """Test suite for ViewerKernel memory management features."""

    def test_current_dset_cache_initialization(self, temp_dir):
        """Test that current_dset_cache is properly initialized."""
        kernel = ViewerKernel(temp_dir)

        assert isinstance(kernel._current_dset_cache, weakref.WeakValueDictionary)
        assert len(kernel._current_dset_cache) == 0

    def test_plot_kwargs_record_initialization(self, temp_dir):
        """Test that plot_kwargs_record is properly initialized."""
        kernel = ViewerKernel(temp_dir)

        assert isinstance(kernel._plot_kwargs_record, dict)
        assert len(kernel._plot_kwargs_record) == 0

    def test_memory_cleanup_threshold(self, temp_dir):
        """Test memory cleanup threshold setting."""
        kernel = ViewerKernel(temp_dir)

        assert kernel._memory_cleanup_threshold == 0.8

        # Test threshold can be modified
        kernel._memory_cleanup_threshold = 0.9
        assert kernel._memory_cleanup_threshold == 0.9


class TestViewerKernelAverageToolbox:
    """Test suite for ViewerKernel average toolbox integration."""

    @patch("xpcsviewer.viewer_kernel._get_module")
    def test_average_toolbox_initialization(self, mock_get_module, temp_dir):
        """Test AverageToolbox initialization."""
        mock_avg_tb_class = Mock()
        mock_avg_tb = Mock()
        mock_avg_tb_class.return_value = mock_avg_tb
        mock_get_module.return_value = mock_avg_tb_class

        kernel = ViewerKernel(temp_dir)

        mock_get_module.assert_called_with("average_toolbox")
        mock_avg_tb_class.assert_called_once_with(temp_dir)
        assert kernel.avg_tb is mock_avg_tb

    @patch("xpcsviewer.viewer_kernel.TableDataModel")
    def test_avg_worker_initialization(self, mock_table_model_class, temp_dir):
        """Test average worker model initialization."""
        mock_model = Mock()
        mock_table_model_class.return_value = mock_model

        kernel = ViewerKernel(temp_dir)

        mock_table_model_class.assert_called_once()
        assert kernel.avg_worker is mock_model
        assert kernel.avg_jid == 0
        assert isinstance(kernel.avg_worker_active, dict)


class TestViewerKernelInheritance:
    """Test suite for ViewerKernel inheritance from FileLocator."""

    def test_inherits_from_file_locator(self, temp_dir):
        """Test that ViewerKernel properly inherits from FileLocator."""
        kernel = ViewerKernel(temp_dir)

        assert isinstance(kernel, FileLocator)
        assert hasattr(kernel, "path")  # FileLocator attribute

    @patch("xpcsviewer.viewer_kernel.FileLocator.__init__")
    def test_calls_super_init(self, mock_super_init, temp_dir):
        """Test that ViewerKernel calls parent __init__."""
        ViewerKernel(temp_dir)

        mock_super_init.assert_called_once_with(temp_dir)


class TestViewerKernelProperties:
    """Test suite for ViewerKernel properties and attributes."""

    def test_path_property(self, temp_dir):
        """Test path property access."""
        kernel = ViewerKernel(temp_dir)

        assert kernel.path == temp_dir

        # Path should be modifiable
        new_path = "/new/test/path"
        kernel.path = new_path
        assert kernel.path == new_path

    def test_statusbar_property(self, temp_dir):
        """Test statusbar property access."""
        mock_statusbar = Mock()
        kernel = ViewerKernel(temp_dir, statusbar=mock_statusbar)

        assert kernel.statusbar is mock_statusbar

        # Statusbar should be modifiable
        new_statusbar = Mock()
        kernel.statusbar = new_statusbar
        assert kernel.statusbar is new_statusbar

    def test_avg_jid_counter(self, temp_dir):
        """Test average job ID counter."""
        kernel = ViewerKernel(temp_dir)

        assert kernel.avg_jid == 0

        # Should be modifiable for job tracking
        kernel.avg_jid = 5
        assert kernel.avg_jid == 5


class TestViewerKernelWeakReferences:
    """Test suite for ViewerKernel weak reference handling."""

    def test_weak_reference_cache_behavior(self, temp_dir):
        """Test that weak reference cache behaves correctly."""
        kernel = ViewerKernel(temp_dir)

        # Create a test object for caching
        class TestObject:
            def __init__(self, value):
                self.value = value

        # Add object to cache
        test_obj = TestObject(42)
        kernel._current_dset_cache["test_key"] = test_obj

        # Verify object is cached
        assert "test_key" in kernel._current_dset_cache
        assert kernel._current_dset_cache["test_key"].value == 42

        # Delete reference - object should be removed from weak cache
        del test_obj

        # Depending on garbage collection timing, the weak reference
        # may or may not be immediately cleared, but accessing it
        # should either work or raise KeyError
        try:
            kernel._current_dset_cache["test_key"]
        except KeyError:
            # Expected behavior when object is garbage collected
            assert "test_key" not in kernel._current_dset_cache


class TestViewerKernelStringRepresentation:
    """Test suite for ViewerKernel string representations."""

    def test_has_docstring(self, temp_dir):
        """Test that ViewerKernel class has proper docstring."""
        ViewerKernel(temp_dir)

        assert ViewerKernel.__doc__ is not None
        assert len(ViewerKernel.__doc__) > 100  # Substantial docstring
        assert "ViewerKernel" in ViewerKernel.__doc__
        assert "Parameters" in ViewerKernel.__doc__
        assert "Attributes" in ViewerKernel.__doc__


class TestViewerKernelPerformance:
    """Test suite for ViewerKernel performance characteristics."""

    def test_initialization_performance(self, temp_dir, performance_timer):
        """Test ViewerKernel initialization performance."""
        performance_timer.start()

        kernel = ViewerKernel(temp_dir)

        elapsed = performance_timer.stop()

        # Initialization should be fast (< 100ms)
        assert elapsed < 0.1
        assert kernel is not None

    def test_meta_reset_performance(self, temp_dir, performance_timer):
        """Test meta reset performance."""
        kernel = ViewerKernel(temp_dir)

        performance_timer.start()
        kernel.reset_meta()
        elapsed = performance_timer.stop()

        # Meta reset should be very fast (< 10ms)
        assert elapsed < 0.01
        assert kernel.meta is not None


class TestViewerKernelEdgeCases:
    """Test suite for ViewerKernel edge cases and error conditions."""

    def test_empty_path_handling(self):
        """Test handling of empty path."""
        with patch("xpcsviewer.viewer_kernel.FileLocator.__init__"):
            kernel = ViewerKernel("")
            assert kernel.path == ""

    def test_none_statusbar_handling(self, temp_dir):
        """Test handling of None statusbar."""
        kernel = ViewerKernel(temp_dir, statusbar=None)
        assert kernel.statusbar is None

    def test_nonexistent_path_handling(self):
        """Test handling of nonexistent path."""
        nonexistent_path = "/nonexistent/path/that/should/not/exist"

        # Should not raise exception during initialization
        with patch("xpcsviewer.viewer_kernel.FileLocator.__init__"):
            kernel = ViewerKernel(nonexistent_path)
            assert kernel.path == nonexistent_path


class TestViewerKernelMemoryIntegration:
    """Test suite for ViewerKernel memory management integration."""

    @patch("xpcsviewer.viewer_kernel.MemoryMonitor")
    def test_memory_monitor_integration(self, mock_memory_monitor, temp_dir):
        """Test integration with MemoryMonitor."""
        ViewerKernel(temp_dir)

        # Kernel should have access to MemoryMonitor through import
        # This tests that the import is successful and accessible
        from xpcsviewer.viewer_kernel import MemoryMonitor as ImportedMemoryMonitor

        assert ImportedMemoryMonitor is not None

    def test_memory_threshold_validation(self, temp_dir):
        """Test memory threshold validation."""
        kernel = ViewerKernel(temp_dir)

        # Default threshold should be valid
        assert 0.0 <= kernel._memory_cleanup_threshold <= 1.0

        # Test setting valid thresholds
        kernel._memory_cleanup_threshold = 0.5
        assert kernel._memory_cleanup_threshold == 0.5

        kernel._memory_cleanup_threshold = 0.95
        assert kernel._memory_cleanup_threshold == 0.95


@pytest.mark.parametrize(
    "path_input,expected_behavior",
    [
        ("/valid/path", "normal"),
        ("relative/path", "normal"),
        ("", "empty"),
        (".", "current_dir"),
        ("/nonexistent", "nonexistent"),
    ],
)
def test_path_variations(path_input, expected_behavior):
    """Test ViewerKernel with various path inputs."""
    with (
        patch("xpcsviewer.viewer_kernel.FileLocator.__init__") as mock_super,
        patch("xpcsviewer.module.average_toolbox.AverageToolbox"),
    ):
        kernel = ViewerKernel(path_input)

        mock_super.assert_called_once_with(path_input)
        assert kernel.path == path_input

        if expected_behavior == "empty":
            assert kernel.path == ""
        elif expected_behavior == "current_dir":
            assert kernel.path == "."
        else:
            assert len(kernel.path) > 0


@pytest.mark.parametrize(
    "statusbar_input",
    [
        None,
        Mock(),
        "invalid_statusbar",
        42,
    ],
)
def test_statusbar_variations(statusbar_input, temp_dir):
    """Test ViewerKernel with various statusbar inputs."""
    with patch("xpcsviewer.viewer_kernel.FileLocator.__init__"):
        kernel = ViewerKernel(temp_dir, statusbar=statusbar_input)

        assert kernel.statusbar is statusbar_input



"""Unit tests for async workers module.

This module provides comprehensive unit tests for async workers,
covering progress management and thread coordination.
"""

import pytest

from xpcsviewer.threading.async_workers import (
    BaseAsyncWorker,
    WorkerResult,
    WorkerSignals,
    WorkerState,
    WorkerStats,
)

# Skip PySide6-dependent tests if not available
pytest_qt = pytest.importorskip("pytestqt", reason="PySide6/Qt tests require pytest-qt")
from PySide6.QtCore import QObject


class TestWorkerState:
    """Test suite for WorkerState enum."""

    def test_worker_state_enum_values(self):
        """Test that WorkerState enum has correct values."""
        assert WorkerState.IDLE.value == "idle"
        assert WorkerState.RUNNING.value == "running"
        assert WorkerState.COMPLETED.value == "completed"
        assert WorkerState.ERROR.value == "error"
        assert WorkerState.CANCELLED.value == "cancelled"

    def test_worker_state_enum_membership(self):
        """Test WorkerState enum membership."""
        assert WorkerState.IDLE in WorkerState
        assert WorkerState.RUNNING in WorkerState
        assert WorkerState.COMPLETED in WorkerState
        assert WorkerState.ERROR in WorkerState
        assert WorkerState.CANCELLED in WorkerState


class TestWorkerResult:
    """Test suite for WorkerResult dataclass."""

    def test_worker_result_creation(self):
        """Test WorkerResult can be created with valid parameters."""
        result = WorkerResult(
            success=True, data={"key": "value"}, error=None, execution_time=1.5
        )

        assert result.success is True
        assert result.data == {"key": "value"}
        assert result.error is None
        assert result.execution_time == 1.5

    def test_worker_result_error_case(self):
        """Test WorkerResult for error cases."""
        error_msg = "Test error"
        result = WorkerResult(
            success=False, data=None, error=error_msg, execution_time=0.1
        )

        assert result.success is False
        assert result.data is None
        assert result.error == error_msg
        assert result.execution_time == 0.1


class TestWorkerStats:
    """Test suite for WorkerStats dataclass."""

    def test_worker_stats_creation(self):
        """Test WorkerStats can be created with valid parameters."""
        stats = WorkerStats(
            total_jobs=10, completed_jobs=8, failed_jobs=1, average_execution_time=2.5
        )

        assert stats.total_jobs == 10
        assert stats.completed_jobs == 8
        assert stats.failed_jobs == 1
        assert stats.average_execution_time == 2.5

    def test_worker_stats_completion_rate(self):
        """Test WorkerStats completion rate calculation."""
        stats = WorkerStats(
            total_jobs=100, completed_jobs=90, failed_jobs=5, average_execution_time=1.0
        )

        # Test completion rate calculation if method exists
        if hasattr(stats, "completion_rate"):
            expected_rate = 90 / 100
            assert stats.completion_rate() == expected_rate


class TestWorkerSignals:
    """Test suite for WorkerSignals Qt object."""

    def test_worker_signals_creation(self):
        """Test WorkerSignals can be created."""
        signals = WorkerSignals()
        assert isinstance(signals, QObject)

    def test_worker_signals_has_required_signals(self):
        """Test WorkerSignals has expected signal attributes."""
        signals = WorkerSignals()

        # Check for common signal attributes (these may vary based on implementation)
        expected_signals = ["progress", "finished", "error", "result"]
        for signal_name in expected_signals:
            if hasattr(signals, signal_name):
                signal_attr = getattr(signals, signal_name)
                # Basic check that it's a signal-like object
                assert hasattr(signal_attr, "connect") or hasattr(signal_attr, "emit")


class TestBaseAsyncWorker:
    """Test suite for BaseAsyncWorker class."""

    def test_base_worker_creation(self):
        """Test BaseAsyncWorker can be instantiated."""
        worker = BaseAsyncWorker()
        assert worker is not None

    def test_base_worker_has_required_methods(self):
        """Test BaseAsyncWorker has expected methods."""
        worker = BaseAsyncWorker()

        # Check for QRunnable interface
        assert hasattr(worker, "run")
        assert callable(worker.run)

    def test_base_worker_state_management(self):
        """Test BaseAsyncWorker state management."""
        worker = BaseAsyncWorker()

        # Check if worker has state management
        if hasattr(worker, "state"):
            # Initial state should be appropriate
            assert worker.state in [WorkerState.IDLE, WorkerState.RUNNING]


@pytest.mark.integration
class TestWorkerIntegration:
    """Integration tests for worker components."""

    def test_worker_result_and_signals_integration(self):
        """Test integration between WorkerResult and WorkerSignals."""
        signals = WorkerSignals()
        result = WorkerResult(
            success=True, data={"test": "data"}, error=None, execution_time=0.5
        )

        # Test that result can be used with signals
        assert result is not None
        assert signals is not None

    def test_worker_stats_and_state_integration(self):
        """Test integration between WorkerStats and WorkerState."""
        stats = WorkerStats(
            total_jobs=5, completed_jobs=3, failed_jobs=1, average_execution_time=1.2
        )

        # Test that stats work with state enum
        assert stats.total_jobs > 0
        assert WorkerState.COMPLETED.value == "completed"


if __name__ == "__main__":
    pytest.main([__file__])



"""SimpleMask window UI setup.

This module provides the setupUi function for SimpleMaskWindow layout.
"""

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QKeySequence
from PySide6.QtWidgets import (
    QButtonGroup,
    QDoubleSpinBox,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSpinBox,
    QSplitter,
    QStatusBar,
    QToolBar,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from xpcsviewer.simplemask.drawing_tools import DRAWING_TOOLS, ERASER_TOOL
from xpcsviewer.simplemask.pyqtgraph_mod import ImageViewROI


def setup_ui(window) -> None:  # noqa: PLR0915 - UI setup requires many statements
    """Set up the SimpleMask window UI.

    Creates a layout with:
    - Toolbar: Drawing tools, mask actions
    - Left panel: ImageView for detector display with ROI support
    - Right panel: Control panels for tools and parameters
    - Status bar for messages

    Args:
        window: SimpleMaskWindow instance to set up
    """
    # Central widget
    central_widget = QWidget()
    window.setCentralWidget(central_widget)

    # Main layout with splitter for resizable panels
    main_layout = QHBoxLayout(central_widget)
    main_layout.setContentsMargins(4, 4, 4, 4)
    main_layout.setSpacing(4)

    splitter = QSplitter(Qt.Orientation.Horizontal)
    main_layout.addWidget(splitter)

    # Left panel: Image display
    left_panel = QWidget()
    left_layout = QVBoxLayout(left_panel)
    left_layout.setContentsMargins(0, 0, 0, 0)

    # PyQtGraph ImageView with ROI support
    window.image_view = ImageViewROI()
    window.image_view.ui.histogram.hide()  # Hide histogram by default
    window.image_view.ui.roiBtn.hide()  # Hide ROI button (we use our own tools)
    window.image_view.ui.menuBtn.hide()  # Hide menu button
    left_layout.addWidget(window.image_view)

    splitter.addWidget(left_panel)

    # Right panel: Controls
    right_panel = QWidget()
    right_layout = QVBoxLayout(right_panel)
    right_layout.setContentsMargins(4, 4, 4, 4)

    # Info label at top
    window.info_label = QLabel("Load data from XPCS Viewer to begin")
    window.info_label.setWordWrap(True)
    right_layout.addWidget(window.info_label)

    # Drawing Tools Group
    drawing_group = QGroupBox("Drawing Tools")
    drawing_layout = QVBoxLayout(drawing_group)

    # Tool buttons
    tool_buttons_layout = QHBoxLayout()
    window.tool_buttons = {}
    window.tool_button_group = QButtonGroup(window)
    window.tool_button_group.setExclusive(True)

    for tool_key, tool in DRAWING_TOOLS.items():
        btn = QToolButton()
        btn.setText(tool.name)
        btn.setToolTip(tool.tooltip)
        btn.setCheckable(True)
        btn.setMinimumWidth(60)
        window.tool_buttons[tool_key] = btn
        window.tool_button_group.addButton(btn)
        tool_buttons_layout.addWidget(btn)

    drawing_layout.addLayout(tool_buttons_layout)

    # Eraser button (separate, not in exclusive group for drawing)
    eraser_layout = QHBoxLayout()
    window.btn_eraser = QToolButton()
    window.btn_eraser.setText(ERASER_TOOL.name)
    window.btn_eraser.setToolTip(ERASER_TOOL.tooltip)
    window.btn_eraser.setCheckable(True)
    window.btn_eraser.setMinimumWidth(60)
    eraser_layout.addWidget(window.btn_eraser)
    eraser_layout.addStretch()
    drawing_layout.addLayout(eraser_layout)

    # Apply drawing button
    apply_layout = QHBoxLayout()
    window.btn_apply_drawing = QPushButton("Apply Drawing")
    window.btn_apply_drawing.setToolTip("Apply all ROIs to create mask")
    apply_layout.addWidget(window.btn_apply_drawing)
    apply_layout.addStretch()
    drawing_layout.addLayout(apply_layout)

    right_layout.addWidget(drawing_group)

    # Mask Actions Group
    mask_group = QGroupBox("Mask Actions")
    mask_layout = QVBoxLayout(mask_group)

    mask_buttons_layout = QHBoxLayout()

    window.btn_undo = QPushButton("Undo")
    window.btn_undo.setToolTip("Undo last mask change (Ctrl+Z)")
    window.btn_undo.setShortcut(QKeySequence.StandardKey.Undo)
    mask_buttons_layout.addWidget(window.btn_undo)

    window.btn_redo = QPushButton("Redo")
    window.btn_redo.setToolTip("Redo mask change (Ctrl+Y)")
    window.btn_redo.setShortcut(QKeySequence.StandardKey.Redo)
    mask_buttons_layout.addWidget(window.btn_redo)

    window.btn_reset = QPushButton("Reset")
    window.btn_reset.setToolTip("Reset mask to initial state")
    mask_buttons_layout.addWidget(window.btn_reset)

    mask_layout.addLayout(mask_buttons_layout)

    # Show/hide mask overlay toggle
    overlay_layout = QHBoxLayout()
    window.btn_toggle_mask = QPushButton("Show Mask")
    window.btn_toggle_mask.setCheckable(True)
    window.btn_toggle_mask.setToolTip("Toggle mask overlay display")
    overlay_layout.addWidget(window.btn_toggle_mask)
    overlay_layout.addStretch()
    mask_layout.addLayout(overlay_layout)

    right_layout.addWidget(mask_group)

    # Geometry Parameters Group
    geometry_group = QGroupBox("Geometry Parameters")
    geometry_layout = QFormLayout(geometry_group)

    # Beam center X
    window.spin_bcx = QDoubleSpinBox()
    window.spin_bcx.setRange(0, 10000)
    window.spin_bcx.setDecimals(2)
    window.spin_bcx.setSuffix(" px")
    window.spin_bcx.setToolTip("Beam center X (column)")
    geometry_layout.addRow("Beam X:", window.spin_bcx)

    # Beam center Y
    window.spin_bcy = QDoubleSpinBox()
    window.spin_bcy.setRange(0, 10000)
    window.spin_bcy.setDecimals(2)
    window.spin_bcy.setSuffix(" px")
    window.spin_bcy.setToolTip("Beam center Y (row)")
    geometry_layout.addRow("Beam Y:", window.spin_bcy)

    # Detector distance
    window.spin_det_dist = QDoubleSpinBox()
    window.spin_det_dist.setRange(1, 100000)
    window.spin_det_dist.setDecimals(1)
    window.spin_det_dist.setSuffix(" mm")
    window.spin_det_dist.setValue(5000.0)
    window.spin_det_dist.setToolTip("Sample-to-detector distance")
    geometry_layout.addRow("Distance:", window.spin_det_dist)

    # Pixel size
    window.spin_pix_dim = QDoubleSpinBox()
    window.spin_pix_dim.setRange(0.001, 10)
    window.spin_pix_dim.setDecimals(4)
    window.spin_pix_dim.setSuffix(" mm")
    window.spin_pix_dim.setValue(0.075)
    window.spin_pix_dim.setToolTip("Pixel dimension")
    geometry_layout.addRow("Pixel Size:", window.spin_pix_dim)

    # X-ray energy
    window.spin_energy = QDoubleSpinBox()
    window.spin_energy.setRange(0.1, 100)
    window.spin_energy.setDecimals(3)
    window.spin_energy.setSuffix(" keV")
    window.spin_energy.setValue(10.0)
    window.spin_energy.setToolTip("X-ray energy")
    geometry_layout.addRow("Energy:", window.spin_energy)

    right_layout.addWidget(geometry_group)

    # Q-Map Actions Group
    qmap_group = QGroupBox("Q-Map")
    qmap_layout = QVBoxLayout(qmap_group)

    # Generate Q-Map button
    qmap_buttons_layout = QHBoxLayout()
    window.btn_generate_qmap = QPushButton("Generate Q-Map")
    window.btn_generate_qmap.setToolTip("Compute Q-map from geometry parameters")
    qmap_buttons_layout.addWidget(window.btn_generate_qmap)
    qmap_buttons_layout.addStretch()
    qmap_layout.addLayout(qmap_buttons_layout)

    # Show Q-Map overlay toggle
    qmap_overlay_layout = QHBoxLayout()
    window.btn_toggle_qmap = QPushButton("Show Q-Map")
    window.btn_toggle_qmap.setCheckable(True)
    window.btn_toggle_qmap.setToolTip("Toggle Q-map overlay display")
    qmap_overlay_layout.addWidget(window.btn_toggle_qmap)
    qmap_overlay_layout.addStretch()
    qmap_layout.addLayout(qmap_overlay_layout)

    right_layout.addWidget(qmap_group)

    # Partition (Q-Binning) Group
    partition_group = QGroupBox("Partition (Q-Binning)")
    partition_layout = QFormLayout(partition_group)

    # Dynamic Q bins
    window.spin_dq_num = QSpinBox()
    window.spin_dq_num.setRange(1, 1000)
    window.spin_dq_num.setValue(10)
    window.spin_dq_num.setToolTip("Number of dynamic Q-bins")
    partition_layout.addRow("Dynamic Q:", window.spin_dq_num)

    # Static Q bins
    window.spin_sq_num = QSpinBox()
    window.spin_sq_num.setRange(1, 1000)
    window.spin_sq_num.setValue(100)
    window.spin_sq_num.setToolTip("Number of static Q-bins")
    partition_layout.addRow("Static Q:", window.spin_sq_num)

    # Dynamic phi bins
    window.spin_dp_num = QSpinBox()
    window.spin_dp_num.setRange(1, 360)
    window.spin_dp_num.setValue(36)
    window.spin_dp_num.setToolTip("Number of dynamic phi-bins")
    partition_layout.addRow("Dynamic φ:", window.spin_dp_num)

    # Static phi bins
    window.spin_sp_num = QSpinBox()
    window.spin_sp_num.setRange(1, 360)
    window.spin_sp_num.setValue(360)
    window.spin_sp_num.setToolTip("Number of static phi-bins")
    partition_layout.addRow("Static φ:", window.spin_sp_num)

    # Compute Partition button
    partition_buttons_layout = QHBoxLayout()
    window.btn_compute_partition = QPushButton("Compute Partition")
    window.btn_compute_partition.setToolTip(
        "Compute Q-space partition from current settings"
    )
    partition_buttons_layout.addWidget(window.btn_compute_partition)
    partition_layout.addRow(partition_buttons_layout)

    # Show partition overlay toggle
    partition_overlay_layout = QHBoxLayout()
    window.btn_toggle_partition = QPushButton("Show Partition")
    window.btn_toggle_partition.setCheckable(True)
    window.btn_toggle_partition.setToolTip("Toggle partition overlay display")
    partition_overlay_layout.addWidget(window.btn_toggle_partition)
    partition_layout.addRow(partition_overlay_layout)

    # Export partition button
    export_layout = QHBoxLayout()
    window.btn_export_partition = QPushButton("Export to Viewer")
    window.btn_export_partition.setToolTip("Export partition to XPCS Viewer")
    export_layout.addWidget(window.btn_export_partition)
    partition_layout.addRow(export_layout)

    right_layout.addWidget(partition_group)

    # Stretch to push remaining controls to bottom
    right_layout.addStretch()

    splitter.addWidget(right_panel)

    # Set splitter sizes (70% image, 30% controls)
    splitter.setSizes([700, 300])
    splitter.setStretchFactor(0, 7)
    splitter.setStretchFactor(1, 3)

    # Store references
    window.splitter = splitter
    window.left_panel = left_panel
    window.right_panel = right_panel
    window.right_layout = right_layout

    # Create menu bar
    _setup_menubar(window)

    # Create toolbar
    _setup_toolbar(window)

    # Status bar
    window.status_bar = QStatusBar()
    window.setStatusBar(window.status_bar)
    window.status_bar.showMessage("Ready")

    # Set window properties
    window.setMinimumSize(1000, 700)
    window.resize(1200, 800)


def _setup_toolbar(window) -> None:
    """Set up the main toolbar with drawing tools and actions.

    Args:
        window: SimpleMaskWindow instance
    """
    toolbar = QToolBar("Main Toolbar")
    toolbar.setMovable(False)
    window.addToolBar(toolbar)

    # Drawing tool actions
    window.tool_actions = {}
    for tool_key, tool in DRAWING_TOOLS.items():
        action = QAction(tool.name, window)
        action.setToolTip(tool.tooltip)
        action.setCheckable(True)
        if tool.shortcut:
            action.setShortcut(QKeySequence(tool.shortcut))
        window.tool_actions[tool_key] = action
        toolbar.addAction(action)

    toolbar.addSeparator()

    # Eraser action
    window.action_eraser = QAction(ERASER_TOOL.name, window)
    window.action_eraser.setToolTip(ERASER_TOOL.tooltip)
    window.action_eraser.setCheckable(True)
    if ERASER_TOOL.shortcut:
        window.action_eraser.setShortcut(QKeySequence(ERASER_TOOL.shortcut))
    toolbar.addAction(window.action_eraser)

    toolbar.addSeparator()

    # Apply drawing action
    window.action_apply = QAction("Apply", window)
    window.action_apply.setToolTip("Apply all drawings to mask")
    toolbar.addAction(window.action_apply)

    toolbar.addSeparator()

    # Mask actions
    window.action_undo = QAction("Undo", window)
    window.action_undo.setShortcut(QKeySequence.StandardKey.Undo)
    window.action_undo.setToolTip("Undo last mask change (Ctrl+Z)")
    toolbar.addAction(window.action_undo)

    window.action_redo = QAction("Redo", window)
    window.action_redo.setShortcut(QKeySequence.StandardKey.Redo)
    window.action_redo.setToolTip("Redo mask change (Ctrl+Y)")
    toolbar.addAction(window.action_redo)

    toolbar.addSeparator()

    # Apply to Viewer button
    window.action_apply_viewer = QAction("Apply to Viewer", window)
    window.action_apply_viewer.setToolTip(
        "Export current mask to XPCS Viewer for analysis"
    )
    toolbar.addAction(window.action_apply_viewer)

    toolbar.addSeparator()

    # Generate Q-Map action
    window.action_generate_qmap = QAction("Generate Q-Map", window)
    window.action_generate_qmap.setToolTip("Compute Q-map from geometry parameters")
    toolbar.addAction(window.action_generate_qmap)

    # Toggle Q-Map overlay
    window.action_toggle_qmap = QAction("Show Q-Map", window)
    window.action_toggle_qmap.setCheckable(True)
    window.action_toggle_qmap.setToolTip("Toggle Q-map overlay display")
    toolbar.addAction(window.action_toggle_qmap)

    toolbar.addSeparator()

    # Toggle mask overlay
    window.action_toggle_mask = QAction("Show Mask", window)
    window.action_toggle_mask.setCheckable(True)
    window.action_toggle_mask.setToolTip("Toggle mask overlay display")
    toolbar.addAction(window.action_toggle_mask)

    window.toolbar = toolbar


def _setup_menubar(window) -> None:
    """Set up the menu bar with File menu.

    Args:
        window: SimpleMaskWindow instance
    """
    menubar = window.menuBar()

    # File menu
    file_menu = menubar.addMenu("&File")

    # Save Mask action
    window.action_save_mask = QAction("&Save Mask...", window)
    window.action_save_mask.setShortcut(QKeySequence.StandardKey.Save)
    window.action_save_mask.setToolTip("Save mask to HDF5 file (Ctrl+S)")
    file_menu.addAction(window.action_save_mask)

    # Load Mask action
    window.action_load_mask = QAction("&Load Mask...", window)
    window.action_load_mask.setShortcut(QKeySequence.StandardKey.Open)
    window.action_load_mask.setToolTip("Load mask from HDF5 file (Ctrl+O)")
    file_menu.addAction(window.action_load_mask)

    file_menu.addSeparator()

    # Apply to Viewer action
    window.action_apply_to_viewer = QAction("&Apply to Viewer", window)
    window.action_apply_to_viewer.setToolTip("Export current mask to XPCS Viewer")
    file_menu.addAction(window.action_apply_to_viewer)

    file_menu.addSeparator()

    # Close action
    window.action_close = QAction("&Close", window)
    window.action_close.setShortcut(QKeySequence.StandardKey.Close)
    window.action_close.setToolTip("Close mask editor window")
    file_menu.addAction(window.action_close)

    window.file_menu = file_menu



key = {
    "nexus": {
        # qmap related
        "mask": "/xpcs/qmap/mask",
        "dqmap": "/xpcs/qmap/dynamic_roi_map",
        "sqmap": "/xpcs/qmap/static_roi_map",
        "dqlist": "/xpcs/qmap/dynamic_v_list_dim0",
        "sqlist": "/xpcs/qmap/static_v_list_dim0",
        "dplist": "/xpcs/qmap/dynamic_v_list_dim1",
        "splist": "/xpcs/qmap/static_v_list_dim1",
        # the beam center is in the qmap is calibrated
        "bcy": "/xpcs/qmap/beam_center_y",
        "bcx": "/xpcs/qmap/beam_center_x",
        "det_dist": "/xpcs/qmap/detector_distance",
        "pixel_size": "/xpcs/qmap/pixel_size",
        "static_index_mapping": "/xpcs/qmap/static_index_mapping",
        "dynamic_index_mapping": "/xpcs/qmap/dynamic_index_mapping",
        "static_num_pts": "/xpcs/qmap/static_num_pts",
        "dynamic_num_pts": "/xpcs/qmap/dynamic_num_pts",
        "start_time": "/entry/start_time",
        "map_names": "/xpcs/qmap/map_names",
        "map_units": "/xpcs/qmap/map_units",
        # multitau related
        "Int_t": "/xpcs/spatial_mean/intensity_vs_time",
        "G2": "/xpcs/multitau/unnormalized_G2",
        "tau": "/xpcs/multitau/delay_list",
        "g2": "/xpcs/multitau/normalized_g2",
        "stride_frame": "/xpcs/multitau/config/stride_frame",
        "avg_frame": "/xpcs/multitau/config/avg_frame",
        "g2_err": "/xpcs/multitau/normalized_g2_err",
        # G2 partial data for stability analysis
        "g2_partial": "/xpcs/multitau/normalized_g2_segments",
        "g2_partial_err": "/xpcs/multitau/normalized_g2_segments_err",
        "g2_partial_labels": "/xpcs/multitau/normalized_g2_segments_labels",
        "saxs_2d": "/xpcs/temporal_mean/scattering_2d",
        "saxs_1d": "/xpcs/temporal_mean/scattering_1d",
        "Iqp": "/xpcs/temporal_mean/scattering_1d_segments",
        "c2_g2_segments": "/xpcs/twotime/normalized_g2_segments",
        "c2_g2": "xpcs/twotime/normalized_g2",
        "c2_delay": "/xpcs/twotime/delay_list",
        "c2_prefix": "/xpcs/twotime/correlation_map",
        "c2_processed_bins": "/xpcs/twotime/processed_bins",
        "c2_stride_frame": "/xpcs/twotime/config/stride_frame",
        "c2_avg_frame": "/xpcs/twotime/config/avg_frame",
        "t0": "/entry/instrument/detector_1/frame_time",
        "t1": "/entry/instrument/detector_1/count_time",
        "X_energy": "/entry/instrument/incident_beam/incident_energy",
        # "ccdx": "/entry/instrument/detector_1/position_x",
        # "ccdy": "/entry/instrument/detector_1/position_y",
        # "ccdx0": "/entry/instrument/detector_1/beam_center_position_x",
        # "ccdy0": "/entry/instrument/detector_1/beam_center_position_y",
        "pix_dim_x": "/entry/instrument/detector_1/x_pixel_size",
        "pix_dim_y": "/entry/instrument/detector_1/y_pixel_size",
        # "bcx": "/entry/instrument/detector_1/beam_center_x",
        # "bcy": "/entry/instrument/detector_1/beam_center_y",
        # "det_dist": "/entry/instrument/detector_1/distance",
    }
}



"""Error handling and edge case testing module for XPCS Toolkit.

This module provides comprehensive testing for error conditions, edge cases,
and failure scenarios to ensure robust operation under all conditions.
"""



#!/usr/bin/env python3
"""
Comprehensive GUI test runner for XPCS Toolkit.

This script provides various test execution modes for GUI testing including
quick validation, full test suite, performance testing, and CI-friendly
headless execution.
"""

import argparse
import os
import subprocess
import sys
from pathlib import Path

# Add the project root to Python path
project_root = Path(__file__).parent.parent.parent
sys.path.insert(0, str(project_root))


def setup_environment():
    """Set up environment variables for GUI testing."""
    # Set Qt platform for headless testing
    os.environ.setdefault("QT_QPA_PLATFORM", "offscreen")

    # Suppress Qt warnings
    os.environ.setdefault("PYXPCS_SUPPRESS_QT_WARNINGS", "1")

    # Set display for headless systems (if needed)
    if not os.environ.get("DISPLAY") and sys.platform.startswith("linux"):
        os.environ["QT_QPA_PLATFORM"] = "offscreen"


def run_pytest(test_args, verbose=True):
    """Run pytest with specified arguments."""
    cmd = ["python", "-m", "pytest"]

    if verbose:
        cmd.append("-v")

    cmd.extend(test_args)

    print(f"Running: {' '.join(cmd)}")
    result = subprocess.run(cmd, check=False, cwd=project_root)
    return result.returncode


def run_quick_tests():
    """Run quick GUI validation tests."""
    print("=== Running Quick GUI Tests ===")

    test_args = [
        "tests/gui_interactive/test_main_window.py",
        "-m",
        "gui and not slow",
        "--tb=short",
        "--maxfail=5",
    ]

    return run_pytest(test_args)


def run_full_test_suite():
    """Run complete GUI test suite."""
    print("=== Running Full GUI Test Suite ===")

    test_args = [
        "tests/gui_interactive/",
        "-m",
        "gui",
        "--tb=line",
        "--continue-on-collection-errors",
    ]

    return run_pytest(test_args)


def run_performance_tests():
    """Run performance and responsiveness tests."""
    print("=== Running Performance Tests ===")

    test_args = [
        "tests/gui_interactive/test_performance.py",
        "-m",
        "performance",
        "--tb=short",
        "-x",  # Stop on first failure for performance tests
    ]

    return run_pytest(test_args)


def run_error_handling_tests():
    """Run error handling and edge case tests."""
    print("=== Running Error Handling Tests ===")

    test_args = [
        "tests/gui_interactive/test_error_handling.py",
        "-m",
        "gui",
        "--tb=short",
    ]

    return run_pytest(test_args)


def run_user_scenario_tests():
    """Run user interaction scenario tests."""
    print("=== Running User Scenario Tests ===")

    test_args = [
        "tests/gui_interactive/test_user_scenarios.py",
        "-m",
        "gui and not slow",
        "--tb=short",
    ]

    return run_pytest(test_args)


def run_ci_tests():
    """Run CI-friendly GUI tests (headless, fast)."""
    print("=== Running CI GUI Tests ===")

    # Ensure headless mode
    os.environ["QT_QPA_PLATFORM"] = "offscreen"

    test_args = [
        "tests/gui_interactive/",
        "-m",
        "gui and not slow and not performance",
        "--tb=no",
        "--quiet",
        "--maxfail=10",
        "--durations=10",
    ]

    return run_pytest(test_args, verbose=False)


def run_coverage_tests():
    """Run GUI tests with coverage reporting."""
    print("=== Running GUI Tests with Coverage ===")

    test_args = [
        "tests/gui_interactive/",
        "-m",
        "gui and not slow",
        "--cov=xpcsviewer",
        "--cov-report=html:htmlcov_gui",
        "--cov-report=term-missing",
        "--tb=short",
    ]

    return run_pytest(test_args)


def run_specific_test_file(test_file):
    """Run a specific test file."""
    print(f"=== Running Specific Test File: {test_file} ===")

    test_path = f"tests/gui_interactive/{test_file}"
    if not Path(project_root / test_path).exists():
        print(f"Error: Test file {test_path} not found")
        return 1

    test_args = [test_path, "-v", "--tb=short"]

    return run_pytest(test_args)


def main():
    """Main test runner function."""
    parser = argparse.ArgumentParser(
        description="XPCS Toolkit GUI Test Runner",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Test Modes:
  quick       - Run basic GUI validation tests (fast)
  full        - Run complete GUI test suite
  performance - Run performance and responsiveness tests
  errors      - Run error handling and edge case tests
  scenarios   - Run user interaction scenario tests
  ci          - Run CI-friendly tests (headless, no slow tests)
  coverage    - Run tests with coverage reporting

Examples:
  python run_gui_tests.py quick
  python run_gui_tests.py full --headless
  python run_gui_tests.py performance
  python run_gui_tests.py test_main_window.py
        """,
    )

    parser.add_argument(
        "mode", nargs="?", default="quick", help="Test mode to run (default: quick)"
    )

    parser.add_argument(
        "--headless", action="store_true", help="Force headless mode (offscreen)"
    )

    parser.add_argument("--verbose", action="store_true", help="Verbose output")

    parser.add_argument("--markers", help="Additional pytest markers to apply")

    args = parser.parse_args()

    # Set up environment
    setup_environment()

    if args.headless:
        os.environ["QT_QPA_PLATFORM"] = "offscreen"

    # Route to appropriate test function
    if args.mode == "quick":
        return run_quick_tests()
    if args.mode == "full":
        return run_full_test_suite()
    if args.mode == "performance":
        return run_performance_tests()
    if args.mode == "errors":
        return run_error_handling_tests()
    if args.mode == "scenarios":
        return run_user_scenario_tests()
    if args.mode == "ci":
        return run_ci_tests()
    if args.mode == "coverage":
        return run_coverage_tests()
    if args.mode.endswith(".py"):
        return run_specific_test_file(args.mode)
    print(f"Unknown test mode: {args.mode}")
    parser.print_help()
    return 1


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)



"""G2 correlation analysis tests for XPCS Toolkit.

This package contains all tests related to G2 correlation function analysis,
fitting algorithms, and validation of correlation analysis results.
"""
